!
! Copyright 2003-2012 Intel Corporation. All Rights Reserved.
!
! The source code contained or described herein and all documents related 
! to the source code ("Material") are owned by Intel Corporation or its 
! suppliers or licensors.  Title to the Material remains with Intel Corporation 
! or its suppliers and licensors.  The Material is protected by worldwide 
! copyright and trade secret laws and treaty provisions.  No part of the 
! Material may be used, copied, reproduced, modified, published, uploaded, 
! posted, transmitted, distributed, or disclosed in any way without Intel's 
! prior express written permission.
!
! No license under any patent, copyright, trade secret or other intellectual
! property right is granted to or conferred upon you by disclosure or delivery
! of the Materials, either expressly, by implication, inducement, estoppel 
! or otherwise.  Any license under such intellectual property rights must
! be express and approved by Intel in writing.
!
!******************************************************************************
! Content:
!      Game of life sample code - serial example.
!      
!******************************************************************************

! maximum size of grid
#define MAXROW  240 
! dependent on size of screen output
#define MAXCOL  600 
! # of generations to compute before displa
#define GENS   1000 

#define DEAD .false.
#define ALIVE .true.

module life
  logical, allocatable, dimension(:,:) :: map0, map1 ! square array holding cells
end module life


program lifeserial
  use life

  integer :: i, gencount=0
  real(kind=8) :: start, time, my_secnds
  character(len=256) :: fName ! file name to read starting grid from
  integer :: length

  fName = ""
  if (command_argument_count() == 1) then
    call get_command_argument(1, fName, length, istatus)
    if(istatus == 0) then
      print *,"FILE NAME: ",trim(fName)
    else
      call Error_Message("Problem reading filename from command line")
    endif
  endif 

  call Initialize(fName)
  start = my_secnds()
  do i=GENS, 1, -1
    gencount = gencount + 1
    if (mod(gencount,2) == 1) then
       call ClearMap(map1)
       call NextGen(map0, map1)
     else
       call ClearMap(map0)
       call NextGen(map1, map0)
      endif
   enddo
   time = my_secnds() - start
   print *,"Time =  ",time
end program lifeserial

subroutine Initialize(fName)
   use life

   character(*):: fName

   ! Put out an initial message and establish the bounds for the grid
   call FindSize()

   ! Laying out the arrays 0 based is a bit inefficient in Fortran, but it
   ! is going to allow the code to match the C code with less contortions.
   allocate(map0(0:MAXROW+1,0:MAXCOL+1), map1(0:MAXROW+1,0:MAXCOL+1), stat=istat)
   if (istat /= 0) call Error_Message("Unable to allocate maps")

   ! Obtain the initial configuration
   call ReadMap(map0, fName)
   ! call WriteMap(map0,0)
   call ClearMap(map1)

end subroutine Initialize

subroutine NextGen(omap, nmap)
   logical, dimension(0:MAXROW+1,0:MAXCOL+1) :: omap, nmap

   integer row, col
   integer ncount
   logical current

   do row = 1, MAXROW
      do col = 1, MAXCOL
         current = omap(row,col)
         ncount = NeighborCount(omap,row,col)
         nmap(row,col) = CellStatus(current,ncount)
      enddo
   enddo
end subroutine NextGen

integer function NeighborCount( map, row, col)         
   logical, dimension(0:MAXROW+1,0:MAXCOL+1) :: map
   integer :: row, col
   integer :: nbrrow, nbrcol ! loop index for row and column of neighbor loops
   integer :: icount
   icount = 0

   do nbrrow = row-1, row+1
      do nbrcol = col-1, col+1
         if (nbrrow /= row .or. nbrcol /= col) then ! skip cell itself
            if (map(nbrrow,nbrcol) == ALIVE) icount = icount + 1
         endif
      enddo
   enddo

   NeighborCount = icount
end function NeighborCount
        


logical function CellStatus(curr, ncount)
    logical :: curr
    integer :: ncount

    logical :: stat

    select case (ncount)
    case (0:1)
       stat = DEAD
    case (2)
       stat = curr
    case (3)
       stat = ALIVE
    case (4:8)
       stat = DEAD
    case default
       call Error_Message("Impossible case in CellStatus.")
    end select

    CellStatus = stat
end function CellStatus

subroutine WriteMap(map, gc)
   logical,dimension(0:MAXROW+1,0:MAXCOL+1) :: map
   integer :: gc
   integer :: row, col, i
   character (len = MAXCOL) :: a

   print *," "
   print *," "
   print *, repeat(" ",MAXCOL/2-7), "GENERATION #", gc
   do row=1,MAXROW
     a = ""
     do col = 1, MAXCOL
       if(map(row,col)) then
         a(col:col) =  '+'
       else
          a(col:col) = '-'
       endif
     enddo
     print *, a
   enddo
end subroutine WriteMap

subroutine FindSize()

   print *,"This program is a simulation of the game of Life"
   print *,"The grid has a size of ", MAXROW, " rows and ", MAXCOL, " columns"
   print *,"Grid coordinates are numbered starting with 1 up to the"
   print *,"number of rows or columns.  Positions outside of the grid"
   print *,"area are incapable of sustaining living cells."
end subroutine FindSize

subroutine ReadMap(map, fName)
  integer row, col
  character(len=*) :: fName
  logical, dimension (0:MAXROW+1,0:MAXCOL+1) :: map
  integer :: good, bad, dup
  good = 0
  bad = 0
  dup = 0

  map = DEAD     ! Set all cells empty, including the hedge

  if (fName == "") then
    print *,"Set up initial cell configuration:"
    print *,"On each line give a pair of coordinates for a living cell."
    print *,"Terminate the list with the special pair 0 0."
    do
      read *, row, col
      if (row == 0 .and. col == 0) exit
      if (row >= 1 .and. row <= MAXROW .and. col >= 1 .and. col <= MAXCOL) then
        map(row,col) = ALIVE
      else
        print *,"Values are not within range."
      end if
    end do
  else      ! read from file
    open(unit = 10, file = trim(fName), status = 'OLD', iostat = istat) 
    if (istat /= 0) then
      call Error_Message("Input file could not be opened")
    endif
    print *,"Reading input data from ",fName
    do 
      read (10,*,IOSTAT= istat) row, col
      if(istat>0) call Error_Message("Read of input file failed")
      if(istat<0) exit
      if (row == 0 .and. col == 0) then
        exit
      else if (row>=1 .and. row<=MAXROW .and. col>=1 .and. col<=MAXCOL) then
        if (map(row,col) == DEAD) then
          good = good + 1
        else  
          dup = dup +1
        endif
        map(row,col) = ALIVE
      else
        bad = bad +1
      endif
    enddo 
    print *,"Done."
    print *, good, "Cells created"
    if (bad > 0) print *, bad, " Cells out of range"
    if (dup > 0) print *, dup, "%d Cells duplicated"
    close(10)
  end if
end subroutine ReadMap

subroutine ClearMap(map)
  logical, dimension(0:MAXROW+1,0:MAXCOL+1) :: map
  map = DEAD
end subroutine ClearMap

real(kind=8) function my_secnds()
  use omp_lib
  my_secnds = omp_get_wtime()
end function my_secnds
