! Copyright 2003-2012 Intel Corporation. All Rights Reserved.
!
! The source code contained or described herein and all documents related 
! to the source code ("Material") are owned by Intel Corporation or its 
! suppliers or licensors.  Title to the Material remains with Intel Corporation 
! or its suppliers and licensors.  The Material is protected by worldwide 
! copyright and trade secret laws and treaty provisions.  No part of the 
! Material may be used, copied, reproduced, modified, published, uploaded, 
! posted, transmitted, distributed, or disclosed in any way without Intel's 
! prior express written permission.
!
! No license under any patent, copyright, trade secret or other intellectual
! property right is granted to or conferred upon you by disclosure or delivery
! of the Materials, either expressly, by implication, inducement, estoppel 
! or otherwise.  Any license under such intellectual property rights must
! be express and approved by Intel in writing.
!
!******************************************************************************
! Content:
!      Monto Carlo sample code - serial version.
!      
!******************************************************************************

#include "timer.f90"

! program to solve Monte Carlo problem 
program mCarlo
  use timer

  implicit none
  include 'mkl.fi'


  integer, parameter ::SEED = 1
  integer (kind=8), parameter :: TIMESTEPS = 1000
  integer, parameter :: PATHS = 1000000

  real (kind=8) :: secs = 0.0
  real (kind=8) :: T, r, sigma, rho, alpha, dt, con1, con2, xmu, var
  real (kind=8) :: payoff, sum0, sum1, ave, sdv
  integer :: mm, nn
  real (kind=8) :: S1, S2
  real (kind=8) :: Y1, Y2
  real (kind=8), dimension(TIMESTEPS) :: rvec0, rvec1

  type(VSL_STREAM_STATE)  :: stream0, stream1
  type(VSL_STREAM_STATE), allocatable ::  stream0Thread(:), stream1Thread(:)
  integer istatus

  call start_timing()

  print '("Monte Carlo simulation of ", I10," paths with ",I10, " timesteps")',&
    PATHS,TIMESTEPS 

  ! set various constants 
  T     = 1.0D0
  r     = 0.05D0
  sigma = 0.1D0
  rho   = 0.5D0
  alpha = sqrt(1.0D0-rho*rho)

  dt    = 1.0/timeSteps

  con1  = 1.0 + r*dt
  con2  = sqrt(dt)*sigma

  xmu = 0.0
  var = 1.0
  payoff = 0.951229424500714 ! exp(-r*T)

  ! initialise VSL random number generator 
  istatus = vslnewstream(stream0, VSL_BRNG_MCG59, SEED)
  ! Create a second instance of the same random stream
  istatus = vslcopystream(stream1, stream0)  
  ! The master stream will have 2 partitions. 0 takes the first, 1 the second
  istatus = vslleapfrogstream( stream0, 0, 2) 
  istatus = vslleapfrogstream( stream1, 1, 2) 

  ! loop over paths 
  sum0 = 0.0
  sum1 = 0.0

  do mm=1, PATHS 

    ! calculate random vector 
    istatus = vdrnggaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2, &
                  stream0, timeSteps, rvec0, xmu, var)
    istatus = vdrnggaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2, &
                  stream1, timeSteps, rvec1, xmu, var)

    ! calculate path 
    S1 = 1.0
    S2 = 1.0

    do nn=1, TIMESTEPS 
      Y1 = rvec0(nn)
      Y2 = rho*Y1 + alpha*rvec1(nn)

      S1 = S1*(con1 + con2*Y1)
      S2 = S2*(con1 + con2*Y2)
    end do

    if ( abs(S1-1.0D0)<0.1D0 .and. abs(S2-1.0D0)<0.1D0 ) then
      sum0 = sum0 + payoff
      sum1 = sum1 + payoff*payoff
    end if
  end do

  ! calculate average and standard deviation
  ave = sum0/PATHS
  sdv = sqrt((sum1/PATHS - ave*ave)/PATHS)
  istatus = vsldeletestream(stream0)
  istatus = vsldeletestream(stream1)

  secs = end_timing()
  print *,"Time Elapsed: ",secs," secs"
  print *,"Average ",ave," Standard deviation of error ",sdv
end program mCarlo


